﻿using Komodex.NETMF.Common;
using Microsoft.SPOT;
using System;
using System.Collections;
using System.Threading;

namespace GoBusExperimental
{
    public class ExternalHub : GoModule, IGoHub
    {
        private readonly Guid _moduleGuid = new Guid(new byte[] { 0x80, 0x70, 0x6E, 0xA9, 0x22, 0xEB, 0x50, 0x4F, 0x92, 0x55, 0x4B, 0xEA, 0x07, 0x85, 0x09, 0x7D });
        public override Guid ModuleGuid { get { return _moduleGuid; } }
        private const int _frameLength = 18;
        private const int _writeRetryCount = 36;
        private const int _readRetryCount = 4;
        // SPI data buffers
        private readonly byte[] _writeFrameBuffer = new byte[_frameLength];
        private readonly byte[] _readFrameBuffer = new byte[_frameLength];

        // Message IDs
        private const byte CMD_READ = (0 << 7);
        private const byte CMD_WRITE = (1 << 7);

        private const byte CMD_CS = 0x01;
        private readonly AutoResetEvent _irqPortSignal = new AutoResetEvent(false);


        protected override void Initialize()
        {
            _goPort.GpioPinMode = GoPortGpioPinMode.IRQ;
            _goPort.OnGpioPinInterrupt += _irqPort_OnInterrupt;

            base.Initialize();
        }

        #region SPI Messages

        private void ClearSPIWriteFrameBuffer()
        {
            for (int i = 0; i < _writeFrameBuffer.Length; i++)
                _writeFrameBuffer[i] = 0;
        }

        private void CalculateCRC()
        {
            _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
        }

        #endregion

        #region IRQ Port

        private void _irqPort_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            _irqPortSignal.Set();
        }

        #endregion

        internal void SetCS(int value)
        {
            ClearSPIWriteFrameBuffer();

            if (value == 0)
            {
                // The hub will deassert the previously selected pin as soon as its own CS pin is asserted.
                // Just write a blank frame to avoid inadvertently sending any data to hub-connected modules.
                _goPort.SPIWrite(_writeFrameBuffer);
            }
            else
            {
                // Set up the message
                if (value != 0)
                {
                    _writeFrameBuffer[0] = 0x80;
                    _writeFrameBuffer[1] = CMD_WRITE | CMD_CS;
                    _writeFrameBuffer[2] = (byte)value;
                    CalculateCRC();
                }
                // Send the message
                _goPort.SPIWrite(_writeFrameBuffer);

                // We can't verify the write because the hub will clear the selected module as soon as its CS pin is asserted.
                // Module writes should be verified on their own, so any necessary verification will be left to module drivers.
            }
        }

        private readonly ArrayList _goPorts = new ArrayList();

        public void AddGoPort(IGoPort goPort)
        {
            goPort.Hub = this;
            _goPorts.Add(goPort);
        }


        public IGoPort[] GoPorts
        {
            get { return (IGoPort[])_goPorts.ToArray(typeof(IGoPort)); }
        }

        public IGoPort GetGoPort(int portNumber)
        {
            throw new NotImplementedException();
        }

        public IEnumerable GetGoPortsForModuleGuid(Guid guid, bool includeBoundPorts)
        {
            IGoPort goPort;
            Guid portGuid;

            for (int i = 0; i < GoPorts.Length; i++)
            {
                goPort = GoPorts[i];
                if (!includeBoundPorts && goPort.BoundModule != null)
                    continue;

                portGuid = goPort.GetConnectedModuleGuid();

                if (guid.Equals(portGuid))
                    yield return goPort;
            }
        }

        public void OnGoPortEnumerating(IGoPort goPort)
        {
        }

        public virtual void OnGoPortEnumerated(IGoPort goPort, bool isBinding)
        {
        }

        public virtual void OnGoPortBinding(IGoPort goPort)
        {
        }

        public virtual void OnGoPortBound(IGoPort goPort)
        {
        }

        public virtual void OnGoPortUnbound(IGoPort goPort)
        {
        }
    }
}
