using System;
using Microsoft.SPOT;
using System.Net.Sockets;
using System.Net;
using System.Threading;
using Microsoft.SPOT.Hardware;

namespace GoBusExperimental
{
    public class NetworkGoHubServer : GoHub
    {
        private Socket _serverSocket;
        private CommunicationsChannel _channel;
        private readonly byte[] _writeBuffer = new byte[1000];
        private readonly byte[] _readBuffer = new byte[1000];

        public NetworkGoHubServer(int port)
        {
            // Start accepting connections
            _serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint localEndPoint = new IPEndPoint(IPAddress.Any, port);
            _serverSocket.Bind(localEndPoint);
            _serverSocket.Listen(Int32.MaxValue);

            // Start our run loop
            Thread runThread = new Thread(RunLoop);
            runThread.Start();
        }

        private void RunLoop()
        {
            while (true)
            {
                // Accept any incoming connection
                Socket clientSocket = _serverSocket.Accept();

                // Close the previous channel
                if (_channel != null)
                    _channel.Stop();

                // Create a new communications channel and start receiving messages
                _channel = new CommunicationsChannel(clientSocket);
                _channel.MessageReceived += CommunicationsChannel_MessageReceived;
                _channel.Start();
            }
        }

        private void CommunicationsChannel_MessageReceived(MessageType messageType, byte[] messageBytes, int offset, int length)
        {
            try
            {
                switch (messageType)
                {
                    case MessageType.SPIWrite:
                        {
                            // Get the port
                            int portID = messageBytes[offset + 0];
                            var port = GetGoPort(portID) as NetworkGoPortServer;
                            if (port == null)
                                return;

                            // Get the SPI speed
                            uint speed = (uint)(messageBytes[offset + 1] << 8);
                            speed |= (uint)(messageBytes[offset + 2] << 0);

                            // Copy the message bytes to our local buffer
                            int byteCount = length - 3;
                            Array.Copy(messageBytes, offset + 3, _writeBuffer, 0, byteCount);

                            // Set SPI parameters and send the message
                            port.SPIClockRate = speed;
                            port.SPIWrite(_writeBuffer, 0, byteCount);
                        }
                        break;

                    case MessageType.SPIWriteRead:
                        {
                            // Get the port
                            int portID = messageBytes[offset + 0];
                            var port = GetGoPort(portID) as NetworkGoPortServer;
                            if (port == null)
                                return;

                            // Get the SPI speed
                            uint speed = (uint)(messageBytes[offset + 1] << 8);
                            speed |= (uint)(messageBytes[offset + 2] << 0);

                            // Copy the message bytes to our local buffer
                            int byteCount = length - 3;
                            Array.Copy(messageBytes, offset + 3, _writeBuffer, 0, byteCount);

                            // Set SPI parameters and send the message
                            port.SPIClockRate = speed;
                            port.SPIWriteRead(_writeBuffer, 0, byteCount, _readBuffer, 0, byteCount, 0);

                            // Now send the bytes we received from the module to the network-connected client
                            _writeBuffer[0] = (byte)portID;
                            Array.Copy(_readBuffer, 0, _writeBuffer, 1, byteCount);
                            _channel.SendMessage(MessageType.SPIWriteReadResponse, _writeBuffer, 0, byteCount + 1);
                        }
                        break;

                    case MessageType.GpioPinMode:
                        {
                            // Get the port
                            int portID = messageBytes[offset + 0];
                            var port = GetGoPort(portID) as NetworkGoPortServer;
                            if (port == null)
                                return;

                            // Get the new pin mode
                            GoPortGpioPinMode pinMode = (GoPortGpioPinMode)messageBytes[offset + 1];

                            // Update the pin mode
                            port.GpioPinMode = pinMode;
                        }
                        break;
                }
            }
            catch { }
        }

        private readonly byte[] _gpioPinInterruptMessage = new byte[1];
        internal void SendGpioPinInterrupt(NetworkGoPortServer goPort)
        {
            // Set up the message
            _gpioPinInterruptMessage[0] = (byte)goPort.ID;

            // Send the message to the client
            _channel.SendMessage(MessageType.GpioPinInterrupt, _gpioPinInterruptMessage, 0, 1);
        }
    }
}
