using System;
using Microsoft.SPOT;
using System.Collections;
using Microsoft.SPOT.Hardware;
using System.Threading;

namespace GoBusExperimental
{
    // Board-specific GoHub implementation

    public class GoHub : GoHubBase, IDisposable
    {
        // Port power on/off delays (in milliseconds)
        private const int PowerOnDelay = 50;
        private const int PowerOffDelay = 50;

        private bool _disposed;

        // Output ports for the shift register on Netduino Go
#pragma warning disable 0169,0649
        private readonly OutputPort _powerShiftRegisterOutputEnable;
        private readonly OutputPort _powerShiftRegisterChipSelect;
        private readonly OutputPort _powerShiftRegisterClock;
        private readonly OutputPort _powerShiftRegisterMosi;
#pragma warning restore 0169,0649

        public GoHub()
        {
#if NETDUINOGO
            _powerShiftRegisterOutputEnable = new OutputPort((Cpu.Pin)10, true);
            _powerShiftRegisterChipSelect = new OutputPort((Cpu.Pin)18, false);
            _powerShiftRegisterClock = new OutputPort((Cpu.Pin)42, false);
            _powerShiftRegisterMosi = new OutputPort((Cpu.Pin)21, false);
#endif
        }

        private readonly ArrayList _goPortPowerStates = new ArrayList();
        private readonly ArrayList _goPortPowerOutput = new ArrayList();
        private readonly ArrayList _goPortLEDs = new ArrayList();

        private bool _powerDownUnboundPorts;
        public virtual bool PowerDownUnboundPorts
        {
            get { return _powerDownUnboundPorts; }
            set
            {
                if (_powerDownUnboundPorts == value)
                    return;
                _powerDownUnboundPorts = value;

                // Update power state for all ports
                var ports = GoPorts;
                for (int i = 0; i < ports.Length; i++)
                {
                    var port = ports[i];

                    // If we're powering down unbound ports and this port has no bound module, turn its power off
                    if (value)
                    {
                        if (port.BoundModule == null)
                            SetPortPowerState(port, false);
                    }
                    // Otherwise, turn the power on (for all ports)
                    else
                    {
                        SetPortPowerState(port, true);
                    }
                }
            }
        }

        public virtual void AddGoPort(IGoPort goPort, Cpu.Pin powerPin, Cpu.Pin ledPin)
        {
            _goPorts.Add(goPort);
            _goPortPowerStates.Add(false);
            if (powerPin >= 0)
                _goPortPowerOutput.Add(new OutputPort(powerPin, false));
            else
                _goPortPowerOutput.Add(null);
            if (ledPin >= 0)
                _goPortLEDs.Add(new OutputPort(ledPin, false));
            else
                _goPortLEDs.Add(null);
            goPort.Hub = this;
            goPort.Initialize();

            // Power on the port if we aren't automatically managing port power
            if (!PowerDownUnboundPorts)
                SetPortPowerState(goPort, true);
        }

        protected virtual void SetPortPowerState(IGoPort goPort, bool state)
        {
            var index = _goPorts.IndexOf(goPort);
            if ((bool)_goPortPowerStates[index] == state)
                return;

            // Set the new power state
            _goPortPowerStates[index] = state;

#if NETDUINOGO
            for (int i = 7; i >= 0; i--)
            {
                bool powerOutput = false;
                if (_goPortPowerStates.Count > i)
                    powerOutput = (bool)_goPortPowerStates[i];
                _powerShiftRegisterMosi.Write(powerOutput);
                _powerShiftRegisterClock.Write(true);
                _powerShiftRegisterClock.Write(false);
            }
            _powerShiftRegisterChipSelect.Write(true);
            _powerShiftRegisterChipSelect.Write(false);
#endif

            var port = (OutputPort)_goPortPowerOutput[index];
            if (port != null)
                port.Write(state);

            // If we're automatically managing port power, add a delay to allow the module to turn on (or off)
            if (PowerDownUnboundPorts)
            {
                if (state)
                    Thread.Sleep(PowerOnDelay);
                else
                    Thread.Sleep(PowerOffDelay);
            }
        }

        protected virtual void SetPortLEDState(IGoPort goPort, bool state)
        {
            var index = _goPorts.IndexOf(goPort);
            var port = (OutputPort)_goPortLEDs[index];
            if (port == null)
                return;

            port.Write(state);
        }

        public override void OnGoPortEnumerating(IGoPort goPort)
        {
            // Set port power if necessary
            if (PowerDownUnboundPorts)
                SetPortPowerState(goPort, true);
        }

        public override void OnGoPortEnumerated(IGoPort goPort, bool isBinding)
        {
            if (!isBinding)
            {
                // Set port power if necessary
                if (PowerDownUnboundPorts)
                    SetPortPowerState(goPort, false);
            }
        }

        public override void OnGoPortBinding(IGoPort goPort)
        {
            // Set port power if necessary
            if (PowerDownUnboundPorts)
                SetPortPowerState(goPort, true);
        }

        public override void OnGoPortBound(IGoPort goPort)
        {
            SetPortLEDState(goPort, true);
        }

        public override void OnGoPortUnbound(IGoPort goPort)
        {
            SetPortLEDState(goPort, false);

            // Set port power if necessary
            if (PowerDownUnboundPorts)
                SetPortPowerState(goPort, false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    foreach (OutputPort port in _goPortPowerOutput)
                    {
                        if (port != null)
                            port.Dispose();
                    }
                    foreach (OutputPort port in _goPortLEDs)
                    {
                        if (port != null)
                            port.Dispose();
                    }
                    if (_powerShiftRegisterOutputEnable != null)
                        _powerShiftRegisterOutputEnable.Dispose();
                    if (_powerShiftRegisterChipSelect != null)
                        _powerShiftRegisterChipSelect.Dispose();
                    if (_powerShiftRegisterClock != null)
                        _powerShiftRegisterClock.Dispose();
                    if (_powerShiftRegisterMosi != null)
                        _powerShiftRegisterMosi.Dispose();
                }
            }
            _disposed = true;
        }

        ~GoHub()
        {
            Dispose(false);
        }
    }
}
