using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

namespace GoBusExperimental
{
    // Most port implementations can safely inherit from GoPortBase since it doesn't contain any hardware-specific code.

    public abstract class GoPortBase : IGoPort
    {
        public virtual int ID { get; set; }
        public virtual IGoHub Hub { get; set; }
        public virtual GoModule BoundModule { get; protected set; }

        public virtual void Initialize()
        {
            // Initialize should be called when this port is added to a hub.
            // This allows the port to initialize its pins as necessary (e.g., setting CS lines high, etc.).
        }

        public Guid GetConnectedModuleGuid()
        {
            return GetConnectedModuleGuid(false);
        }

        protected virtual Guid GetConnectedModuleGuid(bool suppressEnumeratedNotification)
        {
            // If this port is already bound to a module, return that module's GUID
            if (BoundModule != null)
                return BoundModule.ModuleGuid;

            var hub = Hub;

            // Otherwise, send an enumeration request
            // Notify the hub we're about to enumerate (to power up the port if necessary)
            if (hub != null)
                hub.OnGoPortEnumerating(this);

            // Send enumeration request
            var guid = GoBusUtility.GetSPIConnectedModuleGuid(this);
            //Debug.Print("Port " + ID + ": Found module with GUID: " + guid.ToString());

            // Notify the hub we're finished enumerating (to optionally power down the port, etc.)
            if (!suppressEnumeratedNotification && hub != null)
                hub.OnGoPortEnumerated(this, false);

            return guid;
        }

        #region Module Binding

        public virtual void Bind(GoModule module)
        {
            // If this port is already bound to a module, don't attempt to bind it again
            if (BoundModule != null)
                throw new Exception();

            var hub = Hub;

            // Make sure the module connected to this port has a matching GUID
            var connectedGuid = GetConnectedModuleGuid(true);
            if (!module.ModuleGuid.Equals(connectedGuid))
            {
                if (hub != null)
                    hub.OnGoPortEnumerated(this, false);
                throw new Exception();
            }
            // All good, let's go
            if (hub != null)
            {
                hub.OnGoPortEnumerated(this, true);
                hub.OnGoPortBinding(this);
            }

            OnBinding(module);

            // All done!
            BoundModule = module;

            // Notify the hub
            if (hub != null)
                Hub.OnGoPortBound(this);
        }

        protected virtual void OnBinding(GoModule module)
        {
        }

        public virtual void Unbind()
        {
            if (BoundModule == null)
                return;

            OnUnbinding();

            // All done!
            BoundModule = null;

            // Notify the hub
            if (Hub != null)
                Hub.OnGoPortUnbound(this);
        }

        protected virtual void OnUnbinding()
        {
        }

        #endregion

        #region SPI

        public virtual void SPIWrite(byte[] writeBuffer)
        {
            SPIWrite(writeBuffer, 0, writeBuffer.Length);
        }

        public virtual void SPIWrite(byte[] writeBuffer, int writeIndex, int writeLength)
        {
            SPIWriteRead(writeBuffer, writeIndex, writeLength, null, 0, 0, 0);
        }

        public virtual void SPIWriteRead(byte[] writeBuffer, byte[] readBuffer)
        {
            SPIWriteRead(writeBuffer, 0, writeBuffer.Length, readBuffer, 0, readBuffer.Length, 0);
        }

        public abstract void SPIWriteRead(byte[] writeBuffer, int writeIndex, int writeLength, byte[] readBuffer, int readIndex, int readLength, int startReadOffset);

        #endregion

        public abstract GoPortGpioPinMode GpioPinMode { get; set; }
        public abstract bool GpioPinGlitchFilter { get; set; }
        public abstract Port.ResistorMode GpioPinResistorMode { get; set; }
        public abstract Port.InterruptMode GpioPinInterruptMode { get; set; }
        public abstract void GpioWrite(bool value);
        public abstract bool GpioRead();

        public event NativeEventHandler OnGpioPinInterrupt;
        protected virtual void SendOnGpioPinInterrupt(uint data1, uint data2, DateTime time)
        {
            OnGpioPinInterrupt(data1, data2, time);
        }
    }
}
